from __future__ import unicode_literals

from datetime import datetime
from tkinter.messagebox import showerror, askokcancel
from tkinter import *
from threading import Lock, Thread
from inspect import isclass

import os
import logging


class LogHandle(object):
    __file_dir = None
    __base_name = None
    __gui_console = None

    """
    Logging that becomes easier and more uniformed
    """

    def __init__(self, file_dir=None, base_name=None):
        """
        Logging settings not forgotten!

        :param file_dir: File directory of where log files will be saved
        :param base_name: A base filename that log files will be saved as
        """

        self.file_dir = file_dir
        self.base_name = base_name

    @property
    def file_dir(self):
        return self.__file_dir

    @file_dir.setter
    def file_dir(self, file_dir):
        if not isinstance(file_dir, (str, type(None))):
            raise Exception("'file_dir' is missing")
        if file_dir and not os.path.exists(file_dir):
            raise Exception("'file_dir' does not exist")

        self.__file_dir = file_dir

    @property
    def base_name(self):
        return self.__base_name

    @base_name.setter
    def base_name(self, base_name):
        if not isinstance(base_name, (str, type(None))):
            raise Exception("'base_name' is missing")

        self.__base_name = base_name

    @staticmethod
    def __clean_handlers():
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

    def gui_console(self, destroy=False, turn_off=False, gui_obj=None):
        if self.__gui_console and destroy:
            self.__gui_console.destroy()
            self.__gui_console = None
        elif self.__gui_console and turn_off:
            self.__gui_console = None
        elif gui_obj and isclass(gui_obj) and hasattr(gui_obj, "print_gui"):
            self.__gui_console = gui_obj
        else:
            self.__gui_console = GUIConsole()

    def write_to_log(self, message, action='info'):
        """
        Logging becomes more organized!

        :param message: Message that you would like to write to log
        :param action: What kind of action would you like the entry to be marked as?
            (debug, info, warning, error, critical)
        """

        if self.__base_name and self.__file_dir:
            if not message:
                raise Exception("'message' is missing")

            self.__clean_handlers()
            filepath = os.path.join(self.__file_dir,
                                    "{0}_{1}_Log.txt".format(datetime.now().__format__("%Y%m%d"), self.__base_name))

            logging.basicConfig(filename=filepath,
                                level=logging.DEBUG, format=' %(asctime)s - %(levelname)s - %(message)s')

        if self.__gui_console:
            self.__gui_console.print_gui('{0} - {1} - {2}'.format(datetime.now(), action.upper(), message))
        else:
            print('{0} - {1} - {2}'.format(datetime.now(), action.upper(), message))

        if self.__base_name and self.__file_dir:
            if action.lower() == 'debug':
                logging.debug(message)
            elif action.lower() == 'info':
                logging.info(message)
            elif action.lower() == 'warning':
                logging.warning(message)
            elif action.lower() == 'error':
                logging.error(message)
            elif action.lower() == 'critical':
                logging.critical(message)

    def __del__(self):
        logging.shutdown()


class GUIConsole(Toplevel):
    __print_lock = Lock()
    __print_queue = list()
    __console_text = None

    """
        Class to show logging in a GUI window by using Toplevel in tkinter
    """

    def __init__(self):
        Toplevel.__init__(self)
        self.__build()
        self.__start_idle()

    def __start_idle(self):
        self.after(5, self.__on_idle)

    def print_gui(self, msg, sep='\n'):
        with self.__print_lock:
            self.__print_queue.append(str(msg) + sep)

    def __build(self):
        my_frame = Frame(self)
        my_frame.pack()

        self.__console_text = Text(my_frame, bg="black", fg="white")
        self.__console_text.pack()

    def __on_idle(self):
        with self.__print_lock:
            for msg in self.__print_queue:
                self.__console_text.insert(END, msg)
                self.__console_text.see(END)

            self.__print_queue = []

        self.after(5, self.__on_idle)



from __future__ import unicode_literals

from future.moves.queue import LifoQueue, Empty
from threading import Lock
from ..data.error import TransportError

import logging
import os

log = logging.getLogger(__name__)


def close_sql_connections():
    CatchSQLQueue.clear_cache()


class BaseSQLQueue(object):
    CONN_DEFAULT_TIMEOUT = 3
    QUERY_DEFAULT_TIMEOUT = 0
    DEFAULT_CONNECTION_SIZE = 10

    def __init__(self, queue_id, max_pool_size):
        """
        Class to add different SQL engines to a pool

        :param queue_id: [Optional] Queue_ID
        :param max_pool_size: [Optional] Pool_Size for SQL engines
        """
        self.__queue_id = queue_id
        self.__pool_size = max_pool_size
        self.__sql_engine_pool = LifoQueue(maxsize=max_pool_size)
        self.__sql_engine_pool_lock = Lock()
        self.__disabled_pool = list()

    @property
    def pool_list(self):
        """
        :return: Returns list of engines in pool Queue
        """

        if self.__disabled_pool:
            raise Exception("Pool is currently disabled. Unable to view items")
        else:
            return list(self.__sql_engine_pool.queue)

    @property
    def max_pool_size(self):
        """
        :return: Returns Engine pool max size
        """

        if self.__disabled_pool:
            raise Exception("Pool is currently disabled. Unable to get size of pool")
        else:
            return self.__pool_size

    @max_pool_size.setter
    def max_pool_size(self, max_pool_size):
        """
        Sets Max pool size for engine queue

        :param max_pool_size: New max pool size
        """

        if max_pool_size < 1:
            raise ValueError("'max_pool_size' %s is a negative number" % max_pool_size)

        self.__pool_size = max_pool_size

    def create_sql_engine_to_pool(self, sql_config, conn_max_pool_size=DEFAULT_CONNECTION_SIZE,
                                  conn_timeout=CONN_DEFAULT_TIMEOUT, query_timeout=QUERY_DEFAULT_TIMEOUT):
        """
        Creates a new SQL engine and queues it into the engine queue pool

        :param sql_config: SQLConfig instance class
        :param conn_max_pool_size: Max SQL connection pool size
        :param conn_timeout: SQL Connection timeout in seconds
        :param query_timeout: SQL Query timeout in seconds
        """

        from .engine import SQLEngineClass

        if not self.__disabled_pool:
            with self.__sql_engine_pool_lock:
                sql_engine = SQLEngineClass(sql_config=sql_config, conn_max_pool_size=conn_max_pool_size,
                                            conn_timeout=conn_timeout, query_timeout=query_timeout)
                sql_engine.connect()
                sql_engine.engine_sql_class = self

            self.queue_sql_engine_to_pool(sql_engine)
        else:
            log.error('Queue Pool %s: Is in disabled state. Please enable', self.__queue_id)

    def queue_sql_engine_to_pool(self, sql_engine):
        """
        Puts SQLEngine class instance back into the sql queue pool if queue pool has room available

        :param sql_engine: SQLEngine class instance
        """

        from .engine import SQLEngineClass

        if not self.__disabled_pool:
            if not isinstance(sql_engine, SQLEngineClass):
                raise ValueError("'engine' %r is not an instance of SQLEngineClass")

            if sql_engine not in self.__sql_engine_pool.queue:
                with self.__sql_engine_pool_lock:
                    log.debug('Queue Pool %s: Added SQL engine %s to pool', self.__queue_id, sql_engine.engine_id)
                    self.__sql_engine_pool.put(sql_engine, block=False)
        else:
            log.error('Queue Pool %s: Is in disabled state. Please enable', self.__queue_id)

    def pop_sql_engine_from_pool(self):
        """
        Removes sql engine from top of SQL queue pool and returns engine to user

        :return: Returns SQLEngine class instance object
        """

        if not self.__disabled_pool:
            with self.__sql_engine_pool_lock:
                _timeout = 60

                while True:
                    try:
                        log.debug('Queue Pool %s: Waiting for SQL engine', self.__queue_id)
                        sql_engine = self.__sql_engine_pool.get(timeout=_timeout)
                        log.debug('Queue Pool %s: Popped SQL engine %s from pool', self.__queue_id,
                                  sql_engine.engine_id)
                        return sql_engine
                    except Empty:
                        log.debug('Queue Pool %s: No SQL engines available for %s seconds', self.__queue_id, _timeout)
        else:
            log.error('Queue Pool %s: Is in disabled state. Please enable', self.__queue_id)

    def remove_engine_from_pool(self, sql_engine):
        """
        Removes SQLEngine class instance for sql queue pool

        :param sql_engine: SQLEngine class instance
        """

        from .engine import SQLEngineClass

        if not self.__disabled_pool:
            if not isinstance(sql_engine, SQLEngineClass):
                raise ValueError("'engine' %r is not an instance of SQLEngineClass")
            if sql_engine not in self.__sql_engine_pool.queue:
                raise ValueError("'engine' %r is not in the SQL Engine Pool" % sql_engine)

            with self.__sql_engine_pool_lock:

                log.debug('Queue Pool %s: Removed SQL engine %s from pool', self.__queue_id, sql_engine.engine_id)
                self.__sql_engine_pool.queue.remove(sql_engine)
        else:
            log.error('Queue Pool %s: Is in disabled state. Please enable', self.__queue_id)

    def close_pool(self, enable_log=True):
        """
        Closes all connections, cursors from all SQLEngines stored in sql queue pool and empty pool
        """

        with self.__sql_engine_pool_lock:
            if enable_log:
                log.debug('Queue Pool %s: Closing SQL Engine pool', self.__queue_id)

            while True:
                try:
                    self.__sql_engine_pool.get(block=False).close_connections(False, enable_log=enable_log)
                except Empty:
                    break

    def disable_pool(self):
        """
        Closes all connections, cursors from all SQLEngines stored in sql queue pool and disabled sql queue pool
        """

        if not self.__disabled_pool:
            with self.__sql_engine_pool_lock:
                log.debug('Queue Pool %s: Disabling SQL Engine pool', self.__queue_id)

                while True:
                    try:
                        sql_engine = self.__sql_engine_pool.get(block=False)
                        sql_engine.close_connections()
                        self.__disabled_pool.append(sql_engine)
                    except Empty:
                        break
        else:
            log.warning('Queue Pool %s: Is already disabled', self.__queue_id)

    def enable_pool(self):
        """
        Enables sql queue pool and reconnect all SQLEngine instance classes in pool
        """

        if self.__disabled_pool:
            with self.__sql_engine_pool_lock:
                log.debug('Queue Pool %s: Enabling SQL Engine pool', self.__queue_id)

                for sql_engine in self.__disabled_pool:
                    sql_engine.connect()
                    self.__sql_engine_pool.put(sql_engine, block=False)

                self.__disabled_pool = list()
        else:
            log.warning('Queue Pool %s: Is already enabled', self.__queue_id)

    def __del__(self):
        self.close_pool(enable_log=False)

    def __getstate__(self):
        # The pool and lock cannot be pickled
        self.disable_pool()
        state = self.__dict__.copy()
        del state['__sql_engine_pool']
        del state['__sql_engine_pool_lock']
        return state

    def __setstate__(self, state):
        # Restore the pool and lock
        self.__dict__.update(state)
        self.__sql_engine_pool = LifoQueue(maxsize=self.__pool_size)
        self.__sql_engine_pool_lock = Lock()
        self.enable_pool()


class CatchSQLQueue(type):
    _sql_queue_cache = {}
    _sql_queue_cache_lock = Lock()

    def __call__(cls, *args, **kwargs):
        if 'queue_id' in kwargs.keys():
            _sql_queue_cache_key = kwargs['queue_id']
        else:
            _sql_queue_cache_key = id(cls)

        sql = cls._sql_queue_cache.get(_sql_queue_cache_key)

        if isinstance(sql, Exception):
            raise sql
        if sql:
            return sql

        log.debug('Waiting for _sql_queue_cache_lock')

        with cls._sql_queue_cache_lock:
            sql = cls._sql_queue_cache.get(_sql_queue_cache_key)

            if isinstance(sql, Exception):
                raise sql
            if sql:
                return sql

            log.debug("SQLQueue __call__ cache miss. Adding key '%s'", str(_sql_queue_cache_key))

            try:
                sql = super(CatchSQLQueue, cls).__call__(*args, **kwargs)
            except TransportError as e:
                log.warning('Failed to create cached SQLQueue with key %s: %s', _sql_queue_cache_key, e)
                cls._sql_queue_cache[_sql_queue_cache_key] = e
                raise e

            cls._sql_queue_cache[_sql_queue_cache_key] = sql
            return sql

    @classmethod
    def clear_cache(mcs):
        for key, sql_queue in mcs._sql_queue_cache.items():
            if isinstance(sql_queue, Exception):
                continue

            conn_config = key[0]
            log.debug("Conn Config '%s': Closing SQL connections", conn_config)
            sql_queue.close_pool()
            mcs._sql_queue_cache.clear()


class SQLQueue(BaseSQLQueue, metaclass=CatchSQLQueue):
    """
    Creates a SQLQueue pool that is cached for SQLEngine instance classes
    """

    def __init__(self, queue_id=None, max_pool_size=None):
        if not queue_id:
            queue_id = id(self.__class__)

        if not max_pool_size:
            max_pool_size = 10

        super().__init__(queue_id=queue_id, max_pool_size=max_pool_size)



from __future__ import unicode_literals

from independentsoft.msg import Message as Msg
from independentsoft.msg import Recipient, ObjectType, DisplayType, AttachmentMethod, RecipientType, MessageFlag,\
    Attachment, Importance
from exchangelib import FileAttachment, ItemAttachment
from exchangelib import Message as Msg2, Account
from bs4 import BeautifulSoup


class E2M(Msg):
    def __init__(self, ex_lib_item):
        """
        Lets create a .MSG object!

        :param ex_lib_item: Exchangelib message item
        """

        if not isinstance(ex_lib_item, Msg2):
            ValueError("'ex_lib_item' %r is not an instance of ExchangeLib's Message" % ex_lib_item)

        self.__ex_lib_item = ex_lib_item
        self.__msg_attachs = []
        self.__mail_attr()
        self.__rec_proc(self.__ex_lib_item.to_recipients, RecipientType.TO)
        self.__rec_proc(self.__ex_lib_item.cc_recipients, RecipientType.CC)
        self.__mail_attach()
        self.__attach_mail_objs()

        super().__init__()

    def __mail_attr(self):
        self.object_type = ObjectType.MESSAGE
        importance = self.__ex_lib_item.importance
        received_time = self.__ex_lib_item.datetime_received

        if not received_time:
            received_time = self.__ex_lib_item.datetime_created

        if importance == 'Normal':
            self.importance = Importance.NORMAL
        elif importance == 'High':
            self.importance = Importance.HIGH
        elif importance == 'Low':
            self.importance = Importance.LOW
        else:
            self.importance = Importance.NONE

        if received_time:
            received_time = received_time.replace(tzinfo=None)
            self.client_submit_time = received_time
            self.message_delivery_time = received_time

        with self.__ex_lib_item.sender as sender:
            if sender:
                self.reply_to = sender.email_address
                self.in_reply_to = sender.email_address
                self.sender_address_type = sender.routing_type
                self.sender_name = sender.name
                self.sender_search_key = '{0}:{1}\x00'.format(sender.routing_type, sender.email_address).encode()
                self.sender_smtp_address = sender.email_address
                self.sender_email_address = sender.email_address
                self.sent_address_type = sender.routing_type
                self.sent_name = sender.name
                self.sent_search_key = '{0}:{1}\x00'.format(sender.routing_type, sender.email_address).encode()
                self.sent_smtp_address = sender.email_address
                self.sent_email_address = sender.email_address

        if self.__ex_lib_item.display_to:
            self.display_to = self.__ex_lib_item.display_to

        if self.__ex_lib_item.display_cc:
            self.display_cc = self.__ex_lib_item.display_cc

        if self.__ex_lib_item.has_attachments:
            self.has_attachment = self.__ex_lib_item.has_attachments

        self.subject = self.subject
        self.message_flags.append(MessageFlag.HAS_ATTACHMENT)
        self.message_flags.append(MessageFlag.READ)

        if self.__ex_lib_item.text_body:
            self.body_html_text = self.__ex_lib_item.text_body
        elif self.__ex_lib_item.body:
            if bool(BeautifulSoup(self.__ex_lib_item.body, 'html.parser').find()):
                self.body_html_text = self.__ex_lib_item.body
            else:
                self.body_html_text = self.__ex_lib_item.body
        else:
            self.body_html_text = None

    def __rec_proc(self, recs, rec_type):
        if recs:
            for rec in recs:
                self.__pack_rec(rec_type, rec.routing_type, rec.name, rec.email_address)

    def __pack_rec(self, rec_type, rout_type, rec_name, rec_email):
        rec = Recipient()
        rec.address_type = rout_type
        rec.display_type = DisplayType.MAIL_USER
        rec.object_type = ObjectType.MAIL_USER
        rec.display_name = rec_name
        rec.email_address = rec_email
        rec.smtp_address = rec_email
        rec.recipient_type = rec_type
        self.recipients.append(rec)

    def __mail_attach(self):
        if self.__ex_lib_item.attachments:
            for attach in self.__ex_lib_item.attachments:
                if isinstance(attach, FileAttachment):
                    with attach.fp as fp:
                        buffer = fp.read()

                    sub_attach = Attachment(buffer=buffer)
                    sub_attach.file_name = attach.name

                    if attach.is_inline:
                        sub_attach.content_id = attach.content_id
                        sub_attach.method = AttachmentMethod.EMBEDDED_MESSAGE

                    self.attachments.append(sub_attach)
                elif isinstance(attach, ItemAttachment) and isinstance(attach.item, Msg2):
                    self.__msg_attachs.append(ExchangeToMsg(attach.item))

    def __attach_mail_objs(self):
        if self.__msg_attachs:
            for msg in self.__msg_attachs:
                message, subject, received_time = msg.grab_objs()
                sub_attach = Attachment(buffer=message.to_bytes())
                sub_attach.file_name = '%s.msg' % subject
                self.attachments.append(sub_attach)


class ExchangeToMsg(E2M):
    """
    Convert Exchangelib Mail items into .MSG files. Use .save() to save message to filepath
    """
    def __init__(self, ex_lib_item):
        E2M.__init__(self, ex_lib_item)


class Exchange(Account):
    """
    Lets make this a child class of Exchangelib's Account so that we can add renew_session()
    """
    def __init__(self, primary_smtp_address, fullname=None, access_type=None, autodiscover=False, credentials=None,
                 config=None, locale=None, default_timezone=None):
        Account.__init__(self, primary_smtp_address=primary_smtp_address, fullname=fullname, access_type=access_type,
                         autodiscover=autodiscover, credentials=credentials, config=config, locale=locale,
                         default_timezone=default_timezone)

    def renew_session(self):
        """
        Renew e-mail session by retiring old session and starting new session. Great way to keep the connection live
        """

        session = self.protocol.get_session()

        if session:
            self.protocol.retire_session(session)
